module ParseEngine where

import Data.Map hiding (map)
import Text.Regex.Posix
import Grammar
import Result
import Tree
import SrcLoc

import qualified List 




type ParseError = (SrcLoc, String)
type Stream     = (SrcLoc, String)


type Parser a = Stream -> Result a ParseError


type ParseTree = Tree ParseNode

data ParseNode
	= SyntaxNode {
		name :: String,
		start :: SrcLoc,
		end :: SrcLoc
		}
	| TokenNode {
		name :: String,
		start :: SrcLoc,
		end :: SrcLoc,
		value :: String
		}
instance Show ParseNode where
	show (SyntaxNode name s e) = show s ++ "-" ++ show e ++ ": " ++ name
	show (TokenNode name s e val) = show s ++ "-" ++ show e ++ ": " ++ name ++ " " ++ val

matches :: String -> String -> Bool
matches str pattern = str =~ ("^(" ++ pattern ++ ")$")

grab :: String -> String -> String
grab pattern str = str =~ ("^(" ++ pattern ++ ")")

tokenNode :: SrcLoc -> String -> String -> ParseTree
tokenNode loc name token = Tree {
	children = [],
	content = TokenNode {
		name = name,
		start = loc,
		end = loc `shift` (length token - 1),
		value = token
		}}

syntaxNode :: String -> [ParseTree] -> ParseTree
syntaxNode name elems = Tree {
	content = SyntaxNode {
		name = name,
		start = (start . content . head) elems,
		end   = (end   . content . last) elems
		},
	children = elems
	}

--after stream parseNode = 
consume :: String -> Stream -> Stream
consume string (loc, str) = (loc `shift` (length string), drop (length string) str)

loc (l,s) = l

str (l,s) = s

furthest x y = if (loc y) > (loc x)
	then y
	else x

after :: Stream -> ParseTree -> Stream
after (loc, str) tree = (loc', str') where
	loc' = (end (content tree)) `shift` 1
	str' = drop (char loc' - char loc) str



makeGrammarParser :: Grammar -> Parser ParseTree
makeGrammarParser grammar = getParser (mainRule grammar) . skip where

	getParser name = case Data.Map.lookup name parsers of
		Just p -> p
		Nothing -> (\ s -> Failure (loc s, "The grammar does not contain a rule named '" ++ name ++ "'."))
	
	
	parsers :: Data.Map.Map String (Parser ParseTree)
	parsers = fromList $ map makeParser (rules grammar)
	
	
	isBoundary :: Char -> Bool
	isBoundary c = [c] `elem` (delimiters grammar) || [c] `elem` (skipped grammar)
	
	
	skip :: Stream -> Stream
	skip stream = consume taken stream where
		taken = grab skipRegex (str stream)
		skipRegex = concat $ ["(("] ++ List.intersperse ")|(" (skipped grammar)  ++ ["))*"]
	
	
	
	
	
	
	
	makeParser :: (String, ParseRule) -> (String, Parser ParseTree)
	
	-- token parsers
	makeParser (name, Token ttype regexp) = (name, parser) where
		parser :: Parser ParseTree
		parser (loc, "")  = Failure (loc, "end of line found but " ++ name ++ " expected")
		parser (loc, str) = case ttype of
			Bounded -> if token `matches` regexp
				then Success tokenNode'
				else Failure (loc, token ++ " found but " ++ name ++ " expected")
			Greedy -> if token /= ""
				then if isBounded
					then Success tokenNode'
					else Failure (loc `shift` (length token), "delimiter or white space expected after " ++ token)
				else Failure (loc, "failed to parse " ++ name)
			where
				boundedToken :: String
				boundedToken = if [head str] `elem` (delimiters grammar)
					then [head str]
					else takeWhile (not . isBoundary) str
				
				greedyToken :: String
				greedyToken = grab regexp str

				token :: String
				token = case ttype of
					Bounded -> boundedToken
					Greedy  -> greedyToken
				
				isBounded :: Bool
				isBounded = rem == "" || (isBoundary . head) rem where
					rem = drop (length token) str
				
				tokenNode' = tokenNode loc name token
	


	-- syntax rule parsers

	makeParser (name, Syntax rule) = (name, parser) where
		parser stream = case (ruleParser rule) stream of
			Success elems -> Success (syntaxNode name elems)
			Failure f  -> Failure (furthest (loc stream, name ++ " expected") f)

	
	-- Notice that the output is a parser of a list of nodes
	ruleParser :: SyntaxRule -> Parser [ParseTree]
	ruleParser (Either rules) = parser where
		subParsers = map ruleParser rules
		parser stream = alt subParsers err where
			err = (loc stream, "Failed to parse any of the alternatives")
			alt [] err = Failure err
			alt (p:ps) err = case p stream of
				Success s -> Success s
				Failure f -> alt ps (furthest err f)
		
		
	
	ruleParser (Sequence rules) = parser where
		subParsers = map ruleParser rules
		parser stream = seq subParsers stream [] where
			seq [] stream parsed = Success parsed
			seq (p:ps) stream parsed = case p stream of
				Success s -> continueSeq s
				Failure f -> Failure f
				where
					continueSeq s = seq ps stream' parsed' where
						stream' = case parsed' of
							[] -> stream
							_  -> skip $ stream `after` (last parsed')
						parsed' = parsed ++ s
	
	ruleParser (Maybe rule) = parser where
		parser stream = case (ruleParser rule) stream of
			Failure f -> Success []
			Success s -> Success s
	
	ruleParser (ZeroOrMore rule) = parser where
		subParser = ruleParser rule
		parser stream = zom stream []
		zom stream parsed = case subParser stream of
			Success s -> continueZom s
			Failure f -> Success parsed
			where
				continueZom s = zom stream' parsed' where
					stream' = case parsed' of
						[] -> stream
						_  -> skip $ stream `after` (last parsed')
					parsed' = parsed ++ s


	ruleParser (OneOrMore rule) = ruleParser (Sequence [rule, ZeroOrMore rule])


	ruleParser (Rule name) = parser where
		parser stream = case (getParser name) stream of
			Success node -> Success [node]
			Failure f    -> Failure f
	
	ruleParser (DirectToken tok) = parser where
		parser (loc, str) = if tok `List.isPrefixOf` str
			then Success [tokenNode loc "" tok]
			else Failure (loc, tok ++ " expected")
				
	
	
	dummyParsing (loc, str) = Failure (loc, "Not implemented, TODO.")

